#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
from ..constants import TAB


class TabWriter(io.StringIO):
    def __init__(self, *args, tabs: int = 0, **kwargs):
        """
        Utility class for writing tabbed data such as source code
        to the memory buffer in io.StringIO

        @param args: Args to pass to io.StringIO constructor
        @param tabs: Number of starting tabs for this object
        @param kwargs: Kwargs to pass to io.StringIO constructor
        """
        super().__init__(*args, **kwargs)
        self.tabs = tabs

    def write(self, value: str) -> int:
        """
        Writes tabbed data to the internal buffer.

        @param value: String to write.
        @return: Length of data written.
        """
        return super().write(TAB * self.tabs) + super().write(value)

    def writeline(self, value: str) -> int:
        """
        Writes tabbed data and a terminating newline
        to the internal buffer.

        @param value: The string to write.
        @return: Length of data written, including newline character.
        """
        return self.write(value + '\n')

    def indent(self):
        """
        Indents future output.

        @return: None
        """
        self.tabs += 1

    def unindent(self):
        """
        Unindents future output.

        @return: None
        """
        if self.tabs > 0:
            self.tabs -= 1



#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
import argparse
import sys
import os
import os.path
import glob
import clang.cindex
import pxdgen.utils as utils
from pxdgen.utils import TabWriter
from pxdgen.cursors import Namespace, block
from pxdgen.extensions import load_extensions
from colorama import Fore, Style, init as colorama_init
colorama_init()


FLAG_EXTRA_DECLS = "includerefs"
FLAG_IMPORT_ALL = "importall"
FLAG_ERROR_EXIT = "safe"
FLAG_SYS_HEADER = "sys"
FLAG_PARSE_DEFINES = "defines"


SEVERITY = {
    0: '',
    1: Fore.WHITE + "Remark" + Fore.RESET,
    2: Fore.YELLOW + "Warning" + Fore.RESET,
    3: Fore.RED + "Error" + Fore.RESET,
    4: Style.BRIGHT + Fore.RED + "Fatal" + Fore.RESET
}


def px_log(*args, source: str = "PxdGen"):
    source = f"[{source}]"
    print(f"{source:.<15}", *args, sep='', file=sys.stderr)


class PxdGen:
    def __init__(self, program_options: argparse.Namespace):
        """
        Main class for PxdGen supplementary script. CLI implementation.

        @param program_options: Options supplied to CLI.
        """
        dir_mode = os.path.isdir(program_options.header)
        file_mode = os.path.isfile(program_options.header)
        relpath = program_options.relpath or os.getcwd()

        if not dir_mode and not file_mode:
            exit(f"Unable to find input '{program_options.header}' on the file system")

        if not os.path.isdir(relpath):
            exit(f"Unable to find relative path '{relpath}'")

        if dir_mode and not program_options.output:
            request = input("Detected directory without an output location, are you sure you want to print the results to stdout? [y/n]")

            if request.lower() != 'y':
                exit("Use the `-o` flag to specify an output directory")

        if program_options.output and not os.path.isdir(program_options.output):
            os.makedirs(program_options.output)

        if program_options.libs:
            clang.cindex.Config.set_library_path(program_options.libs)

        try:
            self.index = clang.cindex.Index.create()
        except clang.cindex.LibclangError as e:
            msg = str(e)

            if "set_library_path" in msg:
                exit("Unable to find libclang.so. Specify the path to PxdGen with `-L`")

            exit(msg)

        clang_args = list()

        if program_options.include:
            for i in program_options.include:
                clang_args.append("-I")
                clang_args.append(i)

        if program_options.language:
            clang_args += ["-x", program_options.language]

        self.clang_args = clang_args
        self.opts = program_options
        self.dir_mode = dir_mode
        self.file_mode = file_mode
        self.relpath = relpath
        self.flags = set(program_options.flags)

        load_extensions()

    def run(self):
        """
        Run the program with parameters supplied in constructor.

        @return: None.
        """
        if self.opts.recursive and FLAG_EXTRA_DECLS in self.flags:
            exit(f"Recursion (-r) and extra declarations (-f{FLAG_EXTRA_DECLS}) should not be enabled simultaneously")

        to_parse = list()

        # Get the headers to parse
        if self.file_mode:
            to_parse.append(os.path.abspath(self.opts.header))
        elif self.dir_mode:
            for glob_term in ("**/*.h", "**/*.hpp"):
                for header in glob.glob(os.path.join(self.opts.header, glob_term), recursive=True):
                    to_parse.append(os.path.abspath(header))

        valid_headers = set(to_parse)

        if self.opts.verbose:
            px_log(f"PxDGen in {'file mode' if self.file_mode else 'directory mode'} parsing the following header{'s' if len(valid_headers) > 1 else ''}:")

            for h in valid_headers:
                px_log(h)
            px_log()

        # Context which contains a mapping of namespace->IOStream
        # The IOStream has the text for the __init__.pxd of each
        # namespace
        ctx = dict()
        parse_opts = clang.cindex.TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD if FLAG_PARSE_DEFINES in self.flags else 0

        for file in to_parse:
            tu = self.index.parse(file, self.clang_args, options=parse_opts)

            if self.opts.verbose:
                px_log("Parsing ", file)

            for d in tu.diagnostics:
                if d.severity == 0:
                    continue
                if self.opts.verbose and d.severity < 3:
                    px_log(f"{SEVERITY[d.severity]}: {d.spelling}", source="Clang")
                elif d.severity >= 3:
                    px_log(f"{SEVERITY[d.severity]}: {d.spelling}", source="Clang")
                    if FLAG_ERROR_EXIT in self.flags:
                        exit()

            # Find the namespaces for the current file, and its associated cursors
            namespaces = utils.find_namespaces(tu.cursor, valid_headers, self.opts.recursive)
            # The C "top-level" namespace
            namespaces[''] = [tu.cursor]

            for space_name, cursors in namespaces.items():
                pxspace = Namespace(cursors, self.opts.recursive, file, valid_headers)

                if not pxspace.has_declarations:
                    continue

                imports, fwd, body = set(), TabWriter(), TabWriter()

                #  Imports are disabled if extra declarations are defined
                #  Extra declarations are disabled if recursive is enabled
                #  Imports are also disabled if output is directed to a single file/stream
                if FLAG_EXTRA_DECLS in self.flags and not self.opts.recursive:
                    fwd_decls = sorted(pxspace.forward_decls, key=lambda v: len(Namespace._get_all_assoc(v.cursor)))

                    if len(fwd_decls):
                        for line in block(fwd_decls, "toplevel", "cdef extern from *:", False):
                            fwd.writeline(line)
                elif self.opts.output or not self.opts.recursive:
                    for i in pxspace.import_strings(FLAG_IMPORT_ALL in self.flags or self.opts.recursive):
                        imports.add(i)

                for line in pxspace.lines(os.path.relpath(file, self.opts.relpath), FLAG_SYS_HEADER in self.flags):
                    body.writeline(line)

                body.writeline('')

                if self.opts.output:
                    pxd = os.path.splitext(os.path.basename(file))[0] + ".pxd"
                    out_path = os.path.join(self.opts.output, space_name.replace("::", os.path.sep))
                    out_file = os.path.join(out_path, pxd)
                    init = ctx.get(space_name, TabWriter())
                    init_import = os.path.relpath(out_file, self.opts.output).replace(os.path.sep, '.').replace(".pxd", '')
                    init.writeline(f"from {init_import} cimport *")
                    ctx[space_name] = init

                    if not os.path.isdir(out_path):
                        os.makedirs(out_path)

                    stream = open(out_file, 'w')
                else:
                    stream = sys.stdout

                try:
                    for i in sorted(imports):
                        stream.write(i)
                        stream.write('\n')

                    stream.write('\n')

                    if FLAG_EXTRA_DECLS in self.flags:
                        stream.write(fwd.getvalue())
                        stream.write('\n')

                    stream.write(body.getvalue())
                    stream.write('\n')
                finally:
                    if stream is not sys.stdout:
                        stream.close()

        if self.opts.output:
            for space_name in ctx:
                # For C "top-level", do not use __init__.pxd, because
                # The output directory is intended to be added to pxd
                # path
                if not space_name:
                    continue
                out_path = os.path.join(self.opts.output, space_name.replace("::", os.path.sep))

                with open(os.path.join(out_path, "__init__.pxd"), 'w') as out:
                    out.write(ctx[space_name].getvalue())


def main():
    """
    Entry point for PxdGen CLI.

    @return: None
    """
    args = sys.argv[1:]

    argp = argparse.ArgumentParser(description="A tool that converts C/C++ headers to pxd files")
    argp.add_argument("header",
                      help="Path to C/C++ header file or project directory to parse")
    argp.add_argument("-o", "--output",
                      help="Path to output file or directory (defaults to stdout)")
    argp.add_argument("-p", "--relpath",
                      help="Relative path to parse from (defaults to pwd)")
    argp.add_argument("-r", "--recursive",
                      action="store_true",
                      help="Include declarations from all nested headers")
    argp.add_argument("-x", "--language",
                      help="Force Clang to use the specified language for interpretation")
    argp.add_argument("-I", "--include",
                      action="append",
                      help="Add a directory to Clang's include path")
    argp.add_argument("-L", "--libclang-path",
                      dest="libs",
                      help="Specify the path to a directory containing libclang and its dependencies")
    argp.add_argument("-v", "--verbose",
                      action="store_true",
                      help="Print the status of the application to stdout")
    argp.add_argument("-f", "--flag",
                      action="append",
                      dest="flags",
                      default=[],
                      help="Set a flag to further tune the program output")

    opts = argp.parse_args(args)
    proc = PxdGen(opts)
    proc.run()



#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path
import clang.cindex
from typing import List, Dict, Tuple, Callable, Optional
from ..constants import *


def find_namespaces(cursor: clang.cindex.Cursor, valid_headers: set, recursive: bool,
                    **kwargs) -> Dict[str, List[clang.cindex.Cursor]]:
    """
    Finds namespaces, given the top-level cursor of a header file.

    @param cursor: Clang Cursor.
    @param valid_headers: Header whitelist set for filtering.
    @param recursive: Use all namespaces, rather than just the ones in valid_headers.
    @return: Dictionary in the following form:
    {
        "A::B::C": [clang.cindex.Cursor c1, clang.cindex.Cursor c2 ...],
        ...
    }
    """

    def _update(d1, d2):
        for key in d2:
            ns_list = d1.get(key)
            if ns_list is None:
                ns_list = list()
                d1[key] = ns_list
            ns_list += d2[key]

    ret = dict()
    namespaces = list()
    curr_name = kwargs.get("curr_name", '')

    # Add all namespaces under the current cursor
    for child in cursor.get_children():
        if child.location.file is None:
            continue
        add_cond = all((
            child.kind == clang.cindex.CursorKind.NAMESPACE or is_cppclass(child),
            recursive or os.path.abspath(child.location.file.name) in valid_headers
        ))
        if add_cond:
            namespaces.append(child)

    # Recursively process the namespaces added above
    for namespace in namespaces:
        _update(ret, find_namespaces(namespace, valid_headers, recursive, curr_name=curr_name + "::" + namespace.spelling))

    # Add self, if needed
    if cursor.kind in SPACE_KINDS:
        _update(ret, {curr_name.strip("::"): [cursor]})

    return ret


def containing_space(cursor: clang.cindex.Cursor, pred: Callable) -> str:
    """
    Traverse the tree of a cursor and create an address from each
    Node that returns True from pred.

    @param cursor: The root Node.
    @param pred: The predicate to test each Node against.
    @return: A C++-like address.
    """
    parts = list()
    parent = cursor.lexical_parent

    while parent is not None:
        if pred(parent):
            parts.append(parent.spelling)
        parent = parent.lexical_parent

    parts.reverse()

    return "::".join(parts)


def get_cursor_location(cursor: clang.cindex.Cursor) -> str:
    """
    Returns the fully qualified C++ location of a cursor.
    Ex: Baz located in Namespace Foo -> Class Bar (Foo::Bar::Baz)
    would return Foo::Bar

    @param cursor: The cursor to locate.
    @return: Fully qualified C++ location as a string.
    """
    return containing_space(cursor, lambda parent: parent.kind in SPACE_KINDS and not parent.is_inline_namespace())


def get_cursor_namespace(cursor: clang.cindex.Cursor) -> str:
    """
    Returns the C++ namespace of a cursor, disregarding
    parent class/struct types.
    Ex: Baz located in namespace Foo -> Class Bar (Foo::Bar::Baz)
    would return Foo

    @param cursor: The cursor to locate.
    @return: C++ namespace of the cursor.
    """
    return containing_space(cursor, lambda parent: parent.kind == clang.cindex.CursorKind.NAMESPACE and not parent.is_inline_namespace())


def get_cursor_local_access(cursor: clang.cindex.Cursor) -> str:
    """
    Returns the access of a cursor relative to another
    cursor in the same namespace.
    Ex: Fuzz located in namespace Foo -> Class Bar -> Class Baz
    (Foo::Bar::Baz)
    would return Bar::Baz

    @param cursor: The cursor to access.
    @return: C++ access string.
    """
    return containing_space(cursor, lambda parent: parent.kind != clang.cindex.CursorKind.NAMESPACE)


def is_cppclass(cursor: clang.cindex.Cursor) -> bool:
    """
    Given a cursor to a struct/class/class template, returns whether the cursor
    is a C++ class type.

    @param cursor: Clang Cursor.
    @return: Boolean.
    """
    # Handle trivial cases where Clang does the heavy lifting
    if cursor.kind not in STRUCTURED_DATA_KINDS:
        return False
    if cursor.kind in (
            clang.cindex.CursorKind.CLASS_DECL,
            clang.cindex.CursorKind.CLASS_TEMPLATE
    ):
        return True

    # C++ struct decl that is not C-compliant
    for child in cursor.get_children():
        if child.kind == clang.cindex.CursorKind.FIELD_DECL:
            continue
        # There can be anonymous structs and enumerations as fields
        if child.kind in ANON_KINDS and child.is_anonymous():
            continue
        return True

    return False


def is_constructor(cursor: clang.cindex.Cursor) -> bool:
    """
    Whether the cursor represents a constructor. Needed
    because constructors can also be function templates.

    @param cursor: Any Clang cursor.
    @return: Whether the cursor represents a constructor.
    """
    if cursor.kind == clang.cindex.CursorKind.CONSTRUCTOR:
        return True

    if cursor.kind in METHOD_KINDS:
        if cursor.result_type.spelling == "void":
            container = cursor.lexical_parent

            if container is None:
                return False

            func_name = cursor.spelling

            try:
                func_name = func_name[:func_name.index("<")].strip()
            except ValueError:
                pass

            if func_name == container.spelling:
                return True

    return False


def is_forward_decl(cursor: clang.cindex.Cursor) -> bool:
    """
    Returns whether this cursor is a forward declaration.

    @param cursor: Clang cursor.
    @return: bool.
    """
    return cursor.kind in ANON_KINDS and cursor.get_definition() is None


def is_extra_decl(cursor: clang.cindex.Cursor) -> bool:
    """
    Returns whether this is a forward declaration that is
    shadowing the type definition.

    @param cursor: Clang cursor.
    @return: bool.
    """
    d = cursor.get_definition()
    return cursor.kind in ANON_KINDS and d is not None and d != cursor


def is_anonymous(cursor: clang.cindex.Cursor) -> bool:
    """
    Returns whether this struct/union/enum is
    an anonymous/unnamed declaration.

    @param cursor: Clang cursor.
    @return: bool.
    """
    return cursor.kind in ANON_KINDS and cursor.is_anonymous()


def is_typename_unsupported(t: clang.cindex.Type) -> bool:
    """
    Whether a type uses any typename references,
    which are not supported by Cython.

    @param t: Clang type.
    @return: bool.
    """
    ut, _ = get_underlying_type(t)
    return ut.spelling.startswith("typename ") or re.match(RE_DECLTYPE, ut.spelling) or any(
        is_typename_unsupported(ut.get_template_argument_type(i)) for i in range(ut.get_num_template_arguments())
    )


def is_alias_unsupported(cursor: clang.cindex.Cursor) -> bool:
    """
    Determines whether a declaration is based on a
    type that is defined using type aliasing not
    supported by Cython.

    @param cursor: Clang cursor.
    @return: bool.
    """
    while cursor.kind in TYPEDEF_KINDS:
        utt = cursor.underlying_typedef_type

        if is_typename_unsupported(utt):
            return True

        cursor = utt.get_declaration()

    return cursor.kind == clang.cindex.CursorKind.TYPE_ALIAS_TEMPLATE_DECL


def walk_pointer(t: clang.cindex.Type) -> Tuple[int, clang.cindex.Type]:
    """
    Follow a pointer to its underlying type.

    @param t: The pointer to follow. If t is not a pointer, t will be returned.
    @return: The underlying type.
    """
    pointers = 0
    while t.kind == clang.cindex.TypeKind.POINTER:
        t = t.get_pointee()
        pointers += 1

    return pointers, t


def is_function_pointer(ctype: clang.cindex.Type) -> bool:
    """
    Returns whether the type represents a function pointer
    or prototype.

    @param ctype: Any Clang Type object.
    @return: Boolean.
    """
    return walk_pointer(ctype)[1].kind == clang.cindex.TypeKind.FUNCTIONPROTO


def get_function_pointer_return_type(ctype: clang.cindex.Type) -> clang.cindex.Type:
    """
    Gets the return-type of a function pointer or prototype.
    Type is not validated, use is_function_pointer to validate.

    @param ctype: Clang Type object.
    @return: str.
    """
    return walk_pointer(ctype)[1].get_result()


def get_function_pointer_arg_types(ctype: clang.cindex.Type) -> List[clang.cindex.Type]:
    """
    Gets the argument types of a function pointer or prototype.
    Type is not validated, use is_function_pointer to validate.

    @param ctype: Clang Type object.
    @return: list.
    """
    return [arg for arg in walk_pointer(ctype)[1].argument_types()]


def is_function_variadic(cursor: clang.cindex.Cursor) -> bool:
    """
    Determines whether the function represented by
    the supplied cursor is variadic. Uses tokens
    generated by Clang.

    @param cursor: The cursor to check.
    @return: Whether the function was determined to be variadic.
    """
    tokens = [t.spelling for t in cursor.get_tokens()]
    tokens.reverse()

    try:
        return tokens[tokens.index(')') + 1] == "..."
    except ValueError:
        return False


def get_template_params(cursor: clang.cindex.Cursor) -> str:
    """
    Returns the Cython string representing template parameters
    of a Cursor.

    @param cursor: Any Cursor.
    @return: Cython template string, like "[T, U]".
    """

    typenames = list()

    for c in cursor.get_children():
        if c.kind in TEMPLATE_KINDS and c.spelling:
            typenames.append(c.spelling)

    if not len(typenames):
        return ''

    return f"[{', '.join(typenames)}]"


def get_template_params_as_list(cursor: clang.cindex.Cursor) -> List[str]:
    """
    Returns a list containing template parameters
    of a Cursor.

    @param cursor: Any Cursor.
    @return: A Python list like ['T', 'U'].
    """
    return [c.spelling for c in cursor.get_children() if c.kind in TEMPLATE_KINDS and c.spelling]


def get_relative_type_name(importer: clang.cindex.Cursor, importee: clang.cindex.Cursor) -> str:
    """
    How an imported type name appears with respect to another type.

    @param importer: The reference type declaration.
    @param importee: The imported type declaration.
    @return: The string following Cython syntax.
    """
    # Absolute location of importer
    importer_location = get_cursor_location(importer)
    # Absolute location of type being imported
    importee_location = get_cursor_location(importee)
    # Fully qualified C++ name of type being imported
    importee_addr = f"{importee_location}::{importee.spelling}".strip("::")

    # If the location is equal, the imported type can be referred to directly
    # If ignored, just return
    if importer_location == importee_location or importee_addr in IGNORED_IMPORTS:
        return importee.spelling
    # Convenient replacements for special types
    if importee_addr in REPLACED_IMPORTS:
        return REPLACED_IMPORTS[importee_addr]
    # Special handling of libc/libcpp imports that are already defined
    if importee_addr in STD_IMPORTS:
        return importee_addr.replace("::", '_')
    # Top-level namespace, do not process further
    if not importee_location:
        return importee.spelling

    # The containing namespace, not including classes or other spaces
    # In the pxd output, this resolves to a specific file
    importer_namespace = get_cursor_namespace(importer)
    importee_namespace = get_cursor_namespace(importee)
    importee_dot = get_cursor_local_access(importee).split("::")[1:]
    importee_dot.append(importee.spelling)

    # If in the same namespace, we have to access directly or by the containing class
    if importer_namespace == importee_namespace:
        # Not in a class, access directly
        if importee_namespace == importee_location:
            return importee.spelling
        # Create the spelling of class-access
        return '.'.join(importee_dot)
    # Not in the same namespace, use the full reference
    else:
        return (
            importee_namespace.replace("::", '_') +
            '_' + '.'.join(importee_dot)
        )


def get_import_string(importer: clang.cindex.Cursor, importee: clang.cindex.Cursor, import_same_space: bool, default: Optional[str]) -> Optional[str]:
    """
    Calculates an import string given two reference cursors.

    @param importer: The reference type declaration.
    @param importee: The imported type declaration.
    @param import_same_space: Whether types from the
    same namespace should be imported (from separate file).
    @param default: If provided, it serves as an override
    for imports which have no namespace to import from.
    @return: The import string following Cython syntax.
    """
    importer_namespace = get_cursor_namespace(importer)
    importee_namespace = get_cursor_namespace(importee)
    importee_location = get_cursor_location(importee)
    importee_addr = f"{importee_location}::{importee.spelling}".strip("::")

    # Ignored imports are builtin
    if importee_addr in IGNORED_IMPORTS or importee_addr in REPLACED_IMPORTS:
        return None

    importee_dot = get_cursor_local_access(importee).split("::")[1:]
    importee_dot.append(importee.spelling)

    # If in the same file, no import required
    if importer_namespace == importee_namespace:
        if not import_same_space:
            return None

        importer_file = importer.location.file
        importee_file = importee.location.file

        # Same file, no import required
        if importer_file is None or importee_file is None:
            return None
        if importer_file.name == importee_file.name:
            return None

        # If "C-style" and not in C++ namespace, declarations are placed in filename.pxd in output directory
        importee_namespace = importee_namespace or os.path.splitext(os.path.basename(importee_file.name))[0]

        return f"from {importee_namespace.replace('::', '.')} cimport {importee_dot[0]}"

    if not importee_namespace:
        importee_file = importee.location.file

        if importee_file is None:
            return None

        importee_namespace = default or os.path.splitext(os.path.basename(importee_file.name))[0]

        return f"from {importee_namespace} cimport {importee_dot[0]}"

    return "from {} cimport {} as {}".format(
        importee_namespace.replace('::', '.'),
        importee_dot[0],
        importee_namespace.replace('::', '_') + '_' + importee_dot[0]
    )


def full_type_repr(ctype: clang.cindex.Type, ref_cursor: clang.cindex.Cursor) -> str:
    """
    Get the full type string from an existing type.

    @param ctype: The type to convert.
    @param ref_cursor: A reference space to modify output, see example.
    @return: The full type string.
    Examples:
    Case: string<vector<int>>, ref_space = ''
    Returns: std::string<std::vector<int>>
    Case foo::bar, where ref_cursor == foo
    Returns: bar
    """
    expandable = (
        clang.cindex.TypeKind.ELABORATED,
        clang.cindex.TypeKind.UNEXPOSED
    )

    expandable_cursors = (
        clang.cindex.CursorKind.CLASS_DECL,
        clang.cindex.CursorKind.CLASS_TEMPLATE,
        clang.cindex.CursorKind.STRUCT_DECL
    )

    def finalize(subtype: clang.cindex.Type):
        if subtype.spelling in ("bool", "_Bool"):
            return "bint"
        if subtype.spelling in ("const bool", "const _Bool"):
            return "const bint"

        decl = subtype.get_declaration()

        if decl.kind == clang.cindex.CursorKind.NO_DECL_FOUND:
            rep = REPLACED_IMPORTS.get(subtype.spelling)
            return subtype.spelling if rep is None else rep

        return get_relative_type_name(ref_cursor, decl)

    if is_function_pointer(ctype):
        ndim, _ = walk_pointer(ctype)
        result = get_function_pointer_return_type(ctype)
        args = get_function_pointer_arg_types(ctype)
        return f"{full_type_repr(result, ref_cursor)} ({'*' * ndim})({', '.join(full_type_repr(a, ref_cursor) for a in args)})"
    elif ctype.kind == clang.cindex.TypeKind.POINTER:
        ndim, ctype = walk_pointer(ctype)
        return full_type_repr(ctype, ref_cursor) + '*' * ndim
    elif ctype.kind == clang.cindex.TypeKind.LVALUEREFERENCE:
        return full_type_repr(ctype.get_pointee(), ref_cursor) + '&'
    elif ctype.kind == clang.cindex.TypeKind.RVALUEREFERENCE:
        return full_type_repr(ctype.get_pointee(), ref_cursor) + "&&"
    elif ctype.kind == clang.cindex.TypeKind.CONSTANTARRAY:
        return full_type_repr(ctype.get_array_element_type(), ref_cursor) + f"[{ctype.get_array_size()}]"
    elif ctype.kind in (clang.cindex.TypeKind.INCOMPLETEARRAY, clang.cindex.TypeKind.VARIABLEARRAY):
        return full_type_repr(ctype.get_array_element_type(), ref_cursor) + "[]"

    nargs = ctype.get_num_template_arguments()

    if (
        ctype.kind not in expandable or
        ctype.get_declaration().kind not in expandable_cursors or
        nargs <= 0
    ):
        return finalize(ctype)

    params = list()

    for i in range(nargs):
        tmpl_param = ctype.get_template_argument_type(i)

        if not tmpl_param.spelling:
            params.append("void")
        else:
            params.append(full_type_repr(tmpl_param, ref_cursor))

    return f"{finalize(ctype)}<{', '.join(p.strip('*').strip('&') for p in params)}>"


def resolve_typename_type(ctype: clang.cindex.Type, parts: List[str]) -> Optional[clang.cindex.Cursor]:
    """
    Try to resolve C++ types that are resolved with `typename`

    @param ctype: The type to resolve.
    @param parts: Namespace parts eg ["std"].
    @return: Optional cursor, if one is found.
    """
    cur = ctype.translation_unit.cursor.get_children()
    stack = []
    looking = 0

    while True:
        try:
            name = parts[looking]
        except IndexError:
            return None

        for child in cur:
            if child.spelling == name:
                if child.kind in STRUCTURED_DATA_KINDS:
                    return child
                stack.append(cur)
                cur = child.get_children()
                looking += 1
                break
        else:
            if not len(stack):
                return None
            looking -= 1
            cur = stack.pop()


def get_underlying_type(ctype: clang.cindex.Type) -> Tuple[clang.cindex.Type, str]:
    """
    Unwraps the underlying type from pointers, arrays, etc.

    @return: clang.cindex.Type, token tuple
    """
    if ctype.kind == clang.cindex.TypeKind.POINTER:
        ndim, t = walk_pointer(ctype)
        t, tok = get_underlying_type(t)
        return t, tok + ('*' * ndim)
    elif ctype.kind == clang.cindex.TypeKind.LVALUEREFERENCE:
        t, tok = get_underlying_type(ctype.get_pointee())
        return t, tok + '&'
    elif ctype.kind == clang.cindex.TypeKind.RVALUEREFERENCE:
        t, tok = get_underlying_type(ctype.get_pointee())
        return t, tok + "&&"
    elif ctype.kind in (clang.cindex.TypeKind.INCOMPLETEARRAY, clang.cindex.TypeKind.VARIABLEARRAY):
        t, tok = get_underlying_type(ctype.get_array_element_type())
        return t, tok + "[]"
    elif ctype.kind == clang.cindex.TypeKind.CONSTANTARRAY:
        parts = list()

        while ctype.kind == clang.cindex.TypeKind.CONSTANTARRAY:
            parts.append(ctype.get_array_size())
            ctype = ctype.get_array_element_type()

        parts.reverse()
        token = ''

        for t in parts:
            token += f"[{t}]"

        ctype, tok = get_underlying_type(ctype)

        return ctype, tok + token

    return ctype, ''


def strip_type_ids(cursor: clang.cindex.Cursor) -> str:
    """
    Safest method of stripping type strings, but requires a Cursor object
    for introspection. Strips type strings of 'struct', 'enum', and 'union'.

    @param cursor: Clang Cursor.
    @return: str.
    """
    type_spelling = cursor.type.spelling
    replacement = ''

    if cursor.kind == clang.cindex.CursorKind.STRUCT_DECL:
        replacement = "struct"
    elif cursor.kind == clang.cindex.CursorKind.ENUM_DECL:
        replacement = "enum"
    elif cursor.kind == clang.cindex.CursorKind.UNION_DECL:
        replacement = "union"

    if replacement:
        return type_spelling.replace(replacement, '', 1).strip()

    return type_spelling


def strip_all_type_ids(s: str) -> str:
    """
    Deletes all instances of 'struct', 'enum' and 'union' in the type string.
    Can cause issues if the type name has the words in them.

    @param s: Type string to strip.
    @return: Stripped type string.
    """
    return s.replace("struct ", '').replace("enum ", '').replace("union ", '')


def strip_beg_type_ids(s: str) -> str:
    """
    Deletes instances of 'struct', 'enum', and 'union' that exist at the
    beginning of the type string s.

    @param s: Type string to strip.
    @return: Stripped type string.
    """
    ids = ("struct ", "enum ", "union ")

    for i in ids:
        if s.find(i) == 0:
            return s[len(i):]
        #  const params
        const_term = "const " + i
        if s.find(const_term) == 0:
            return s.replace(const_term, "const ")

    return s


def convert_dialect(s: str) -> str:
    """
    Converts C++ dialect string to Cython dialect
    string. Replaces template delimiters and removes
    some names that are valid in C(++) but not Cython

    @param s: String to convert
    @return: Converted string
    """
    throws = "throw("

    # First templates
    ret = s.replace('<', '[').replace('>', ']')

    # Replace exception information
    tloc = ret.find(throws)
    if tloc != -1:
        eb = ret.index(')', tloc)
        ret = ret.replace(ret[tloc:eb+1], "except +")
    else:
        ret = ret.replace("noexcept", '')

    ret = ret.replace("restrict ", '').replace("volatile ", '').replace("typename ", '').replace("::", '.')

    return ret



#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import clang.cindex
from . import utils
from .constants import *
from typing import Optional, Generator, Set, Any, Tuple, List


def specialize(cursor: clang.cindex.Cursor) -> CCursor:
    """
    Determine what abstracted class defined in this
    module to use for a specific cursor.

    @param cursor : A cursor to convert to a class from this module.
    @return       : The converted type.
    """
    if cursor.kind in BASIC_DATA_KINDS:
        return DataType(cursor)
    elif utils.is_constructor(cursor):
        return Constructor(cursor)
    elif cursor.kind in FUNCTION_KINDS:
        return Function(cursor)
    elif cursor.kind == clang.cindex.CursorKind.ENUM_DECL:
        return Enumeration(cursor)
    elif cursor.kind == clang.cindex.CursorKind.UNION_DECL:
        return Union(cursor)
    elif cursor.kind in TYPEDEF_KINDS:
        return Typedef(cursor)
    elif cursor.kind in STRUCTURED_DATA_KINDS:
        return Struct(cursor)
    elif cursor.kind == clang.cindex.CursorKind.MACRO_DEFINITION:
        return Macro(cursor)
    else:
        return CCursor(cursor)


def block(children: List[CCursor], name: str, header: str, restrict: bool) -> Generator[str, None, None]:
    """
    Iterate over the lines of a block type.

    @param children : The children of the block.
    @param name     : The name of the block. Used for naming anonymous types.
    @param header   : The header of the block to yield.
    @param restrict : A restricted space such as a C struct that must
                      have restricted declarations.
    @return         : Generator[str]
    """
    forward_decls = list()
    anonymous_decls = list()

    for child in children:
        if child.forward:
            forward_decls.append(child)
        elif child.anonymous:
            anonymous_decls.append(child)
        for assoc in CCursor._base_assoc_types(child.cursor):
            if assoc.forward:
                forward_decls.append(assoc)
            elif assoc.anonymous:
                anonymous_decls.append(assoc)

    forward_decls = list(set(forward_decls))
    anonymous_decls = list(set(anonymous_decls))
    prefix = '' if restrict else TAB

    if not restrict:
        yield header

    for i, cursor in enumerate(anonymous_decls):
        for line in cursor.lines(name=f"pxdgen_anon_{name}_{i}"):
            yield prefix + line
    for decl in forward_decls:
        for line in decl.lines():
            yield prefix + line

    if restrict:
        yield header

    typedefs = list()
    declared_typedefs = list()
    count = 0

    for child in children:
        if child.anonymous or child.forward:
            continue
        if child.cursor.kind in ANON_KINDS and not child.name:
            typedefs.append(child)
            continue
        count += 1
        if isinstance(child, DataType):
            ut, token = utils.get_underlying_type(child.cursor.type)
            decl = ut.get_declaration()

            try:
                ind = anonymous_decls.index(decl)
            except ValueError:
                ind = -1

            if ind != -1:
                i = token.find('[')

                if i != -1:
                    suffix = token[i:]
                    token = token[:i]
                else:
                    suffix = ''

                yield TAB + f"pxdgen_anon_{name}_{ind}{token} {child.name}{suffix}"
                continue
        elif isinstance(child, Typedef):
            ut, token = utils.get_underlying_type(child.cursor.underlying_typedef_type)
            utd = ut.get_declaration()

            try:
                ind = typedefs.index(utd)
            except ValueError:
                ind = -1

            if ind != -1:
                td = typedefs.pop(ind)

                for line in td.lines(name=child.name, typedef=True):
                    yield TAB + line

                declared_typedefs.append((td, child.name))
                continue

            found = False

            for decl, name in declared_typedefs:
                if decl == utd:
                    yield TAB + f"ctypedef {name}{token} {child.name}"
                    found = True
                    break

            if found:
                continue

        for i, line in enumerate(child.lines()):
            commented = line.strip().startswith('#')
            if restrict and isinstance(child, Function) and child.is_static_method:
                yield TAB + ("#  " if commented else '') + "@staticmethod"
            yield TAB + line

        if i == 0 and commented:
            count -= 1

    if not count:
        yield TAB + "pass"


class CCursor:
    def __init__(self, cursor: clang.cindex.Cursor):
        self.cursor = cursor
        self._address = "::".join((self.location, self.name)).strip("::")

    def __iter__(self):
        return self.cursor.get_children()

    def __eq__(self, other: Any):
        if isinstance(other, CCursor):
            return other.cursor == self.cursor
        if isinstance(other, clang.cindex.Cursor):
            return self.cursor == other
        raise NotImplementedError(f"CCursor not comparable to type {type(other)}")

    def __hash__(self) -> int:
        return self._address.__hash__()

    @property
    def is_space(self):
        return self.cursor.kind in SPACE_KINDS

    @property
    def is_functional(self):
        return self.cursor.kind in FUNCTION_KINDS

    @property
    def is_structured_data(self):
        return self.cursor.kind in STRUCTURED_DATA_KINDS

    @property
    def is_static(self) -> bool:
        """
        Whether the storage for this element is static.

        @return: Boolean.
        """
        return self.cursor.storage_class == clang.cindex.StorageClass.STATIC

    @property
    def parent(self) -> Optional[CCursor]:
        p = self.cursor.lexical_parent
        return None if p is None else specialize(p)

    @property
    def visible(self) -> bool:
        return self.cursor.access_specifier == clang.cindex.AccessSpecifier.PUBLIC

    @property
    def file(self) -> Optional[str]:
        try:
            return os.path.abspath(self.cursor.location.file.name)
        except AttributeError:
            return None

    @property
    def anonymous(self) -> bool:
        return utils.is_anonymous(self.cursor)

    @property
    def forward(self) -> bool:
        return utils.is_forward_decl(self.cursor)

    @property
    def name(self) -> str:
        return self.cursor.spelling

    @property
    def display(self) -> str:
        return self.cursor.displayname

    @property
    def namespace(self) -> str:
        return utils.get_cursor_namespace(self.cursor)

    @property
    def location(self) -> str:
        return utils.get_cursor_location(self.cursor)

    @property
    def address(self) -> str:
        return self._address

    @property
    def associated_types(self) -> Set[CCursor]:
        """
        Get the associated types within this cursor.

        Example:
        Case: std::vector<std::map<std::string, int>>
        Result: [vector, map, string] (clang.cindex.Cursors)
        Remarks: Int is not included as it is built-in.
        @return: List of cursors containing the definition of each type.
        """
        return CCursor._base_assoc_types(self.cursor)

    @property
    def unsupported(self) -> bool:
        """
        Whether this declaration uses a C++ construct
        unsupported by Cython.

        @return: bool.
        """
        unsupported_types = (clang.cindex.TypeKind.BLOCKPOINTER,)

        if self.cursor.type.kind in unsupported_types:
            return True

        for child in self.cursor.get_children():
            if child.kind in TYPE_REFS:
                d = child.get_definition() or child.type.get_declaration()

                if utils.is_alias_unsupported(d):
                    return True

        return utils.is_typename_unsupported(self.cursor.type)

    @staticmethod
    def _base_assoc_types(cursor: clang.cindex.Cursor) -> Set[CCursor]:
        result = set()

        for child in cursor.get_children():
            if child.kind in TYPE_REFS:
                cdef = child.get_definition()
                if cdef is not None:
                    result.add(specialize(cdef))
                else:
                    cdef = child.type.get_declaration()
                    if cdef.kind != clang.cindex.CursorKind.NO_DECL_FOUND:
                        result.add(specialize(cdef))

        return result

    def lines(self, **kwargs) -> Generator[str, None, None]:
        raise NotImplementedError(f"Abstract method `lines` called in CCursor abstract class with cursor {self.cursor.kind}:{self.cursor.spelling}")


class DataType(CCursor):
    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents a simple variable or constant declaration,
        function parameter, struct/class/union field, etc.

        @param cursor: Associated Clang cursor.
        """
        super().__init__(cursor)

    @property
    def associated_types(self) -> Set[CCursor]:
        """
        Override from base class to support
        function pointers.
        """
        result = super().associated_types

        if self.is_function_pointer:
            for child in self.cursor.get_children():
                if child.kind == clang.cindex.CursorKind.PARM_DECL:
                    result.update(DataType(child).associated_types)

        return result

    @property
    def unsupported(self) -> bool:
        if self.is_function_pointer:
            for child in self.cursor.get_children():
                if child.kind == clang.cindex.CursorKind.PARM_DECL:
                    if DataType(child).unsupported:
                        return True

        return super().unsupported

    @property
    def is_function_pointer(self) -> bool:
        return utils.is_function_pointer(self.cursor.type)

    @property
    def declaration(self) -> str:
        """
        The declaration of this Member in Cython syntax.

        @return: Cython syntax str.
        """
        if self.unsupported:
            return f"#  {self.cursor.type.spelling} {self.name}"

        if self.is_function_pointer:
            return self._function_ptr_declaration

        ut, token = utils.get_underlying_type(self.cursor.type)
        utd = ut.get_declaration()

        # If the anonymous declaration has not already been handled, default
        if utils.is_anonymous(utd):
            # Pointer coersion is easier for void* than array of chars
            if token.count('*') == len(token):
                typename = f"void{token}"
            else:
                typename = f"char[{self.cursor.type.get_size()}]"
        else:
            typename = self.typename

        # Clang refers to arrys in int[20] syntax
        # This block extracts the array size portion,
        # and is saved to suffix so that the end
        # result looks like int data[20]
        suffix = ''
        ob = typename.find('[')

        if ob != -1:
            suffix = typename[ob:]
            typename = typename[:ob].strip()

        return f"{utils.convert_dialect(typename)} {self.name}{suffix}"

    @property
    def typename(self) -> str:
        return utils.full_type_repr(self.cursor.type, self.cursor)

    @property
    def _function_ptr_declaration(self) -> str:
        """
        Special declaration for function pointers.

        @return: str.
        """
        typename = self.typename
        i = typename.index(')')

        return utils.convert_dialect(typename[:i] + self.name + typename[i:])

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        The lines of this DataType.

        @param kwargs: None.
        @return: Generator of lines.
        """
        yield self.declaration


class Macro(CCursor):
    def __init__(self, cursor):
        super().__init__(cursor)

    def lines(self, **kwargs) -> Generator[str, None, None]:
        macro_type = "const int"
        tokens = [t.spelling for t in self.cursor.get_tokens()]
        func = self.cursor.is_macro_function()
        i = 1 if not func else (tokens.index(')') + 1)
        macro_def = ''.join(tokens[i:])

        if re.fullmatch(RE_CPP_INT, macro_def):
            macro_type = "const long"
        elif re.fullmatch(RE_CPP_FLOAT, macro_def):
            macro_type = "const double"

        yield f"{macro_type} {tokens[0]}{'(...)' if func else ''}"


class Function(CCursor):
    CYTHON_UNSUPPORTED = {
        "operator+=",
        "operator-=",
        "operator^=",
        "operator&=",
        "operator|=",
        "operator->",
        # Python name collisions
        "is",
        "global",
    }

    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents a functional type declaration.

        @param cursor: Clang cursor.
        """
        super().__init__(cursor)
        self._tmpl_params = utils.get_template_params(cursor)
        # get_arguments() yields nothing for function templates
        # testing cursor.kind manually for now to handle both cases
        self._args = [DataType(child) for child in cursor.get_children() if child.kind == clang.cindex.CursorKind.PARM_DECL]

    @property
    def is_static_method(self) -> bool:
        """
        Whether this function is a static method.

        @return: Boolean.
        """
        return self.cursor.is_static_method()

    @property
    def associated_types(self) -> Set[CCursor]:
        """
        Associated types for this function.

        @return: Set[CCursor]
        """
        result = super().associated_types

        for arg in self._args:
            result.update(arg.associated_types)

        return result

    @property
    def unsupported(self) -> bool:
        return (
            any(arg.unsupported for arg in self._args) or
            utils.is_typename_unsupported(self.cursor.result_type) or
            super().unsupported
        )

    @property
    def first_optional_arg_index(self) -> int:
        """
        Gets the first optional argument index for
        this function. Returns len(args) if there
        are no default parameters.

        @return: First optional argument index.
        """
        n = len(self._args)

        for i, arg in enumerate(self._args):
            if any(child.kind == clang.cindex.CursorKind.UNEXPOSED_EXPR for child in arg):
                n = i
                break

        return n

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        Generates the lines for this function. Functions with
        default parameter values will yield multiple overloads.

        @param kwargs: No kwargs to specify.
        @return: Generator over lines.
        """
        comment = (
            self.unsupported or
            self.cursor.spelling in Function.CYTHON_UNSUPPORTED or
            self.cursor.spelling.startswith('operator""')
        )
        n = self.first_optional_arg_index
        restype = utils.convert_dialect(utils.full_type_repr(self.cursor.result_type, self.cursor))

        for i in range(n, len(self._args) + 1):
            args = self._argument_declarations(i)
            prefix = "#  " if comment else ''

            if utils.is_function_pointer(self.cursor.result_type):
                i = restype.index(')')
                yield f"{prefix}{restype[:i]}{self.name}{self._tmpl_params}({', '.join(args)}){restype[i:]}"
            else:
                yield f"{prefix}{restype} {self.name}{self._tmpl_params}({', '.join(args)})"

    def _argument_declarations(self, nargs: int) -> Generator[str, None, None]:
        """
        Returns whether there was an error and the Cython argument
        declarations of this function.
        """

        for mem in self._args[:nargs]:
            # `=*` syntax seems to only work for cdef functions defined in pyx files
            # suffix = "=*" if any(child.kind == clang.cindex.CursorKind.UNEXPOSED_EXPR for child in mem) else ''
            yield utils.convert_dialect(mem.typename)

        # Variadic argument does not get represented in children or get_arguments()
        # So utils function uses tokens
        if utils.is_function_variadic(self.cursor):
            yield "..."


class Constructor(Function):
    def __init__(self, cursor: clang.cindex.Cursor):
        super().__init__(cursor)

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        Lines for this constructor.

        @param kwargs: None for this type.
        @return: Generator over lines.
        """
        spelling = self.name
        restype = "void " if self._tmpl_params else ''
        n = self.first_optional_arg_index

        try:
            spelling = spelling[:spelling.index('<')]
        except ValueError:
            pass

        for i in range(n, len(self._args) + 1):
            args = self._argument_declarations(i)

            yield ("#  " if self.unsupported else '') + f"{restype}{spelling}{self._tmpl_params}({', '.join(args)})"


class Enumeration(CCursor):
    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents an enum, given an enum Cursor.

        @param cursor: Clang cursor.
        """
        super().__init__(cursor)

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        The lines of this enumeration.

        @param kwargs: typedef : bool - should this use the ctypedef syntax.
                       name    : str  - The name to use for this declaration.
        @return: Generator over lines.
        """
        yield Enumeration.cython_header(kwargs.get("typedef", False), kwargs.get("name", self.name))

        for line in self.members():
            yield TAB + line

    def members(self) -> Generator[str, None, None]:
        """
        Iterator over the members of this enum.

        @return: Generator[str, None, None]
        """
        gen = 0
        for kind, spelling, enum_value in Enumeration.iterate(self.cursor):
            if kind == clang.cindex.CursorKind.ENUM_CONSTANT_DECL:
                gen += 1
                yield f"{spelling} = {enum_value}"

        if not gen:
            yield "pass"

    @staticmethod
    def iterate(cursor: clang.cindex.Cursor) -> Generator[Tuple[clang.cindex.CursorKind, str, int], None, None]:
        for child in cursor.get_children():
            yield child.kind, child.spelling, child.enum_value

    @staticmethod
    def cython_header(typedef: bool, name: str) -> str:
        """
        The Cython header declaration for this enum.

        @param typedef: Whether this is a typedef'd enum.
        @param name: The name to use for this declaration.
        @return: str.
        """
        return f"{'ctypedef ' if typedef else ''}enum {name}:"


class Union(CCursor):
    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents a union, given a union Cursor.

        @param cursor: Clang cursor.
        """
        super().__init__(cursor)
        self._children = list()

        for child in cursor.get_children():
            if (child.kind == clang.cindex.CursorKind.FIELD_DECL and child.spelling) or child.kind in ANON_KINDS:
                self._children.append(specialize(child))

    @property
    def associated_types(self) -> Set[CCursor]:
        result = set()

        for child in self._children:
            result.update(child.associated_types)

        return result

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        The lines of this Union.

        @param kwargs: typedef : bool - Define using ctypedef syntax.
                       name    : str  - The name to use for this declaration.
        @return: Generator[str]
        """
        name = kwargs.get("name", self.name)
        for line in block(
            self._children,
            name,
            Union.cython_header(kwargs.get("typedef", False), name),
            True
        ):
            yield line

    @staticmethod
    def cython_header(typedef: bool, name: str) -> str:
        """
        The Cython header declaration for this union.

        @param typedef: Whether this is a typedef'd union.
        @param name: The name to use for this declaration.
        @return: str.
        """
        return f"{'ctypedef ' if typedef else ''}union {name}:"


class Typedef(CCursor):
    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents a typedef, given the correct Clang Cursor.

        @param cursor: Clang typedef Cursor.
        """
        super().__init__(cursor)

    @property
    def underlying_type(self) -> clang.cindex.Type:
        """
        Unwraps the underlying type from pointers, arrays, etc.

        @return: The underlying type. A better form of `underlying_typedef_type`.
        """
        return utils.get_underlying_type(self.cursor.underlying_typedef_type)[0]

    @property
    def unsupported(self) -> bool:
        return (
            utils.is_typename_unsupported(self.cursor.underlying_typedef_type) or
            super().unsupported
        )

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        Cython lines for this typedef.

        @param kwargs: None.
        @return: str.
        """
        utt = self.cursor.underlying_typedef_type

        if "__builtin_va_list" in (self.name, utt.spelling):
            yield f"ctypedef void* {self.name}"
            return
        if self.unsupported:
            yield f"#  ctypedef {utt.spelling} {self.name}"
            return

        result = utils.convert_dialect(utils.full_type_repr(utt, self.cursor))

        if utils.is_function_pointer(utt):
            # A function prototype can be typedefed but is not an lvalue reference
            if utt.kind == clang.cindex.TypeKind.FUNCTIONPROTO:
                yield "ctypedef " + result.replace("()", self.name, 1)
            else:
                i = result.index(')')
                yield "ctypedef " + result[:i] + self.name + result[i:]
            return

        ut, token = utils.get_underlying_type(utt)
        utd = ut.get_declaration()

        # If typedef of another type has not been handled from higher level context
        if utd.kind in ANON_KINDS and not result.strip():
            for line in specialize(utd).lines(name=self.name, typedef=True):
                yield line
            return

        yield f"ctypedef {result} {self.name}"


class Struct(CCursor):
    # Types yielded from members property
    INSTANCE_TYPES = (
        clang.cindex.CursorKind.FIELD_DECL,
        clang.cindex.CursorKind.CONSTRUCTOR,
        clang.cindex.CursorKind.CXX_METHOD,
        clang.cindex.CursorKind.FUNCTION_TEMPLATE,
        clang.cindex.CursorKind.TYPEDEF_DECL,
        clang.cindex.CursorKind.TYPE_ALIAS_DECL,
        clang.cindex.CursorKind.ENUM_DECL,
        clang.cindex.CursorKind.CLASS_DECL,
        clang.cindex.CursorKind.STRUCT_DECL,
        clang.cindex.CursorKind.CLASS_TEMPLATE,
        clang.cindex.CursorKind.UNION_DECL
    )

    def __init__(self, cursor: clang.cindex.Cursor):
        """
        Represents a Cython struct/cppclass, given the correct
        Clang Cursor.

        @param cursor: Clang struct/cppclass Cursor.
        """
        super().__init__(cursor)
        self._is_cppclass = utils.is_cppclass(cursor)
        self._children = list()
        self._tmpl_params = utils.get_template_params(cursor)

        for child in cursor.get_children():
            if (
                    child.access_specifier == clang.cindex.AccessSpecifier.PRIVATE or
                    child.kind not in Struct.INSTANCE_TYPES or
                    utils.is_extra_decl(child) or
                    (child.kind not in ANON_KINDS and not child.spelling)
            ):
                continue
            self._children.append(specialize(child))

    @property
    def is_cppclass(self) -> bool:
        """
        Whether this is a C++ class or POD C struct.

        @return: Boolean.
        """
        return self._is_cppclass

    @property
    def associated_types(self) -> Set[CCursor]:
        result = set()

        for child in self._children:
            result.update(child.associated_types)

        return result

    def lines(self, **kwargs) -> Generator[str, None, None]:
        """
        Lines of this class/structure.

        @param kwargs: typedef : bool - Should this use ctypedef syntax.
                       name    : str  - The name to use for this declaration.
        @return: Generator[str]
        """
        name = kwargs.get("name", self.name)

        for line in block(
            self._children,
            name,
            self.cython_header(kwargs.get("typedef", False), name),
            True
        ):
            yield line

    def cython_header(self, typedef: bool, name: str) -> str:
        """
        The Cython header for this struct/class.

        @param typedef: Whether this should be printed as a typedef.
        @param name: The name to use for this header.
        @return: str.
        """
        return "{}{} {}{}:".format(
            "ctypedef " if typedef and not self.is_cppclass else '',
            "cppclass" if self.is_cppclass else "struct",
            name,
            self._tmpl_params
        )


class Namespace:
    def __init__(self, cursors: list, recursive: bool, main_header: str,  valid_headers: set, *_):
        """
        Represents a Cython namespace declaration, given the following parameters.

        @param cursors: A list of cursors associated with this namespace.
        @param recursive: Whether this namespace should declare children from other headers, recursively.
        @param main_header: The header file to accept declarations from.
        @param valid_headers: The set of headers being parsed.
        """
        self.cursors = [CCursor(c) for c in cursors]
        self.cpp_name = self.cursors[0].address if cursors[0].kind in SPACE_KINDS else ''
        self.recursive = recursive
        self.main_header = main_header
        self.valid_headers = valid_headers
        self.children = list()
        self.class_space = all(utils.is_cppclass(c) for c in cursors)

        for cursor in self.cursors:
            self.children += filter(self._child_filter, cursor)

    @property
    def has_declarations(self) -> bool:
        """
        Whether this space has any valid declarations.

        @return: Boolean.
        """
        return len(self.children) > 0

    @property
    def forward_decls(self) -> Set[CCursor]:
        """
        Empty declarations needed to resolve types that
        have not been included.

        @return: Set of CCursors of type declarations that
        need to be defined.
        """
        assert not self.recursive, "Forward declarations should not be used when recursion is enabled"

        result = set()

        for child in self.children:
            for t in Namespace._get_all_assoc(child):
                addr = t.address
                if (
                    t.file != self.main_header and
                    addr not in IGNORED_IMPORTS and
                    addr not in REPLACED_IMPORTS
                ):
                    result.add(t)

        return result

    def lines(self, rel_header_path: str, system_header: bool) -> Generator[str, None, None]:
        """
        Generator over the lines of this namespace.

        @param rel_header_path: The relative header path to the header where this namespace is defined.
        @param system_header: Whether angled brackets should be added to the header name.
        @return: Generator[str]
        """
        children = [specialize(c) for c in self.children]
        name = self.cursors[0].address if self.cursors[0].cursor.kind in SPACE_KINDS else 'toplevel'

        for line in block(
            children,
            name,
            self.cython_header(rel_header_path, system_header),
            False
        ):
            yield line

    def cython_header(self, rel_header_path: str, system_header: bool) -> str:
        """
        The Cython header for this namespace.

        @param rel_header_path: Header path relative to the C compiler's include path.
        @param system_header: Whether angled brackets should be added to the header name.
        @return: The Cythonized header for this namespace.
        """
        header_name = rel_header_path.replace(os.path.sep, '/')
        header_name = f"<{header_name}>" if system_header else header_name
        base = f"cdef extern from \"{header_name}\""
        namespace = f" namespace \"{self.cpp_name}\":" if self.cpp_name else ':'

        return base + namespace

    def import_strings(self, include_all: bool) -> Set[str]:
        """
        Import strings required for this namespace.

        @param include_all: Whether all imports from unsubmitted headers
        should be inculded.
        @return: A set of import strings.
        """
        result = set()

        for child in self.children:
            for t in specialize(child).associated_types:

                if t.cursor.kind in TEMPLATE_KINDS:
                    continue

                if t.file not in self.valid_headers:
                    # Handle if import should be done via libc/libcpp
                    stdpath = STD_IMPORTS.get(t.address)

                    if stdpath is not None:
                        result.add(f"from {stdpath} cimport {t.name} as {t.address.replace('::', '_')}")
                        continue

                    if not include_all:
                        continue

                default = os.path.splitext(os.path.basename(self.main_header))[0] if self.recursive else None
                res = utils.get_import_string(child, t.cursor, not self.recursive, default)

                if res is not None:
                    result.add(res)

        return result

    def _child_filter(self, child: clang.cindex.Cursor):
        """
        Filters out invalid/invisible child cursors of this namespace.
        """
        if child.access_specifier == clang.cindex.AccessSpecifier.PRIVATE:
            return False
        if self.class_space and child.kind in Struct.INSTANCE_TYPES:
            return False
        if utils.is_extra_decl(child):
            return False
        if type(specialize(child)) is CCursor:
            return False
        try:
            return (
                self.recursive or
                os.path.abspath(child.location.file.name) == self.main_header
            )
        except AttributeError:
            return False

    @staticmethod
    def _get_all_assoc(cursor: clang.cindex.Cursor) -> Set[CCursor]:
        result = set()
        stack = [cursor]

        while len(stack):
            current = stack.pop()

            for child in current.get_children():
                stack.append(child)

                if (child.kind in TYPE_REFS or
                        (current.kind == clang.cindex.CursorKind.TYPEDEF_DECL and
                         child.kind in STRUCTURED_DATA_KINDS and
                         child.spelling)):
                    decl = child.get_definition()

                    if decl is None:
                        continue

                    spec = specialize(decl)

                    # Check to avoid parsing the same type multiple times.
                    # Greatly reduces the time required
                    if spec not in result:
                        stack.append(decl)
                        result.add(spec)

        return result



#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import clang.cindex

kinds = clang.cindex.CursorKind
TAB_LENGTH = 4
TAB = ' ' * TAB_LENGTH
RE_DECLTYPE = re.compile("decltype\(.+\)")
RE_CPP_INCLUDE = re.compile(" *#include *[<\"].+\.h(pp)?[>\"] *")
RE_CPP_INT = re.compile("\d{1,20}")
RE_CPP_FLOAT = re.compile("\d+(\.\d+)?[fF]?")

SPACE_KINDS = (
    kinds.STRUCT_DECL,
    kinds.CLASS_DECL,
    kinds.CLASS_TEMPLATE,
    kinds.NAMESPACE
)
TEMPLATE_KINDS = (
    kinds.TEMPLATE_TYPE_PARAMETER,
    kinds.TEMPLATE_NON_TYPE_PARAMETER
)
FUNCTION_KINDS = (
    kinds.FUNCTION_DECL,
    kinds.FUNCTION_TEMPLATE,
    kinds.CXX_METHOD
)
METHOD_KINDS = (
    kinds.FUNCTION_TEMPLATE,
    kinds.CXX_METHOD
)
STATIC_FUNCTION_KINDS = (
    kinds.FUNCTION_DECL,
    kinds.FUNCTION_TEMPLATE
)
STRUCTURED_DATA_KINDS = (
    kinds.STRUCT_DECL,
    kinds.CLASS_DECL,
    kinds.CLASS_TEMPLATE
)
# Anything 100% able to be represented by the DataType class
BASIC_DATA_KINDS = (
    kinds.PARM_DECL,
    kinds.VAR_DECL,
    kinds.FIELD_DECL
)
ANON_KINDS = STRUCTURED_DATA_KINDS + (
    kinds.ENUM_DECL,
    kinds.UNION_DECL
)
TYPE_REFS = (
    kinds.TYPE_REF,
    kinds.TEMPLATE_REF
)
TYPEDEF_KINDS = (
    kinds.TYPEDEF_DECL,
    kinds.TYPE_ALIAS_DECL
)
IGNORED_IMPORTS = {
    "size_t",
    "ptrdiff_t",
    "wchar_t"
}
REPLACED_IMPORTS = {
    "std::size_t": "size_t"
}
STD_IMPORTS = {
    "lconv": "libc.locale",
    "jmp_buf": "libc.setjmp",
    "sigjmp_buf": "libc.setjmp",
    "sig_handler_t": "libc.signal",
    "sig_atomic_t": "libc.signal",
    "int8_t": "libc.stdint",
    "int16_t": "libc.stdint",
    "int32_t": "libc.stdint",
    "int64_t": "libc.stdint",
    "uint8_t": "libc.stdint",
    "uint16_t": "libc.stdint",
    "uint32_t": "libc.stdint",
    "uint64_t": "libc.stdint",
    "intptr_t": "libc.stdint",
    "uintptr_t": "libc.stdint",
    "intmax_t": "libc.stdint",
    "uintmax_t": "libc.stdint",
    "FILE": "libc.stdio",
    "fpos_t": "libc.stdio",
    "div_t": "libc.stdlib",
    "ldiv_t": "libc.stdlib",
    "lldiv_t": "libc.stdlib",
    "clock_t": "libc.time",
    "time_t": "libc.time",
    "tm": "libc.time",
    "std::complex": "libcpp.complex",
    "std::deque": "libcpp.deque",
    "std::list": "libcpp.list",
    "std::map": "libcpp.map",
    "std::unique_ptr": "libcpp.memory",
    "std::shared_ptr": "libcpp.memory",
    "std::weak_ptr": "libcpp.memory",
    "std::queue": "libcpp.queue",
    "std::priority_queue": "libcpp.queue",
    "std::set": "libcpp.set",
    "std::multiset": "libcpp.multiset",
    "std::stack": "libcpp.stack",
    "std::string": "libcpp.string",
    "std::unordered_map": "libcpp.unordered_map",
    "std::unordered_set": "libcpp.unordered_set",
    "std::unordered_multiset": "libcpp.unordered_set",
    "std::pair": "libcpp.pair",  # Also in libcpp.utility
    "std::vector": "libcpp.vector"
}

del kinds



#    PxdGen - A C/C++ header conversion tool
#    Copyright (C) 2021  Eric Rowley

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup


with open("./README.md", 'r') as file:
    long_d = file.read()


setup(
    name="pxdgen",
    version="1.1.0",
    url="https://github.com/earowley/pxdgen",
    py_modules=["_pxdgen"],
    packages=["pxdgen", "pxdgen.utils"],
    package_dir={
        "": "scripts",
        "pxdgen": "src/pxdgen"
    },
    entry_points={
        "console_scripts": ["pxdgen = _pxdgen:main"]
    },
    license="GNU GPLv3",
    author="Eric Rowley",
    author_email="earowley23@gmail.com",
    description="A tool for converting one or more C/C++ headers to Cython pxd files.",
    long_description=long_d,
    long_description_content_type="text/markdown",
    install_requires=[
        "clang",
        "colorama"
    ],
    python_requires="~=3.6"
)
